import numpy as np
from warnings import warn
import pandas as pd
from scipy.optimize import fsolve

from pyMelt.meltingcolumn_classes import MeltingColumn_1D


class mantle:
    """
    The mantle class consists of one or more lithology classes, in a particular proportion. The
    mantle class contains the methods used for doing melting calculations and for calculating the
    properties of a heterogeneous mantle.

    Parameters
    ----------
    lithologies :    list of lithology objects
        A list of lithology instances.
    proportions :    list or numpy.array of floats
        The mass ratios of the lithologies, doesn't need to be normalised.
    names :  list of strings or None, default: None
        The names of the lithologies. If False, default names will be chosen.

    Attributes
    ----------
    number_lithologies : int
        the number of lithologies in the mantle class
    CP :     list of floats
        the heat capacities of the lithologies
    alphaf : list of floats
        the thermal expansion coefficients of the melts produced by each lithology.
    alphas : list of floats
        the thermal expansion coefficients of each lithology
    rhof :   list of floats
        the densities of the melts produced by each lithology
    rhos :   list of floats
        the densities of each lithology
    DeltaS : list of floats
        the entropy change on melting of each lithology.

    """
    def __init__(self, lithologies, proportions, names=None):
        self.lithologies = lithologies
        if isinstance(proportions, list):
            proportions = np.array(proportions)
        self.proportions = proportions/sum(proportions)
        self.number_lithologies = np.shape(self.lithologies)[0]
        if names is None:
            names = list()
            for i in range(self.number_lithologies):
                names.append('default '+str(i))
        self.names = names

        self.CP = np.zeros(self.number_lithologies)
        self.alphaf = np.zeros(self.number_lithologies)
        self.alphas = np.zeros(self.number_lithologies)
        self.rhof = np.zeros(self.number_lithologies)
        self.rhos = np.zeros(self.number_lithologies)
        self.DeltaS = np.zeros(self.number_lithologies)
        for i in range(self.number_lithologies):
            self.CP[i] = self.lithologies[i].CP
            self.alphaf[i] = self.lithologies[i].alphaf
            self.alphas[i] = self.lithologies[i].alphas
            self.rhof[i] = self.lithologies[i].rhof
            self.rhos[i] = self.lithologies[i].rhos
            self.DeltaS[i] = self.lithologies[i].DeltaS

    def bulk_properties(self, P=None, T=None):
        """
            Calculates the bulk thermodynamic properties of the solid or partially
            molten mantle.

            Parameters
            ----------
            P : float or None, default: None
                The pressure of interest. If None, the properties of the solid mantle will be
                returned.
            T : float or None, default: None
                The temperature of interest. If None, the properties of the solid mantle will be
                returned.

            Returns
            -------
            dict
                The bulk alpha, CP and rho for the mantle at the given P and T, labelled
                as such.
        """
        F = np.zeros(self.number_lithologies)
        if T is not None:
            for i in range(self.number_lithologies):
                F[i] = self.lithologies[i].F(P, T)
        alpha = sum(self.proportions*self.alphaf*F + self.proportions*self.alphas*(1-F))
        CP = sum(self.proportions*self.CP)
        rho = sum(self.proportions*self.rhof*F + self.proportions*self.rhos*(1-F))

        return {'alpha': alpha, 'CP': CP, 'rho': rho}

    def solidus_intersection(self, Tp):
        """
        Finds the pressure at which each lithology's solidus will be intersected,
        assuming the mantle follows the solid adiabat up until that point.

        Parameters
        ----------
        Tp : float
            The mantle potential temperature in degC.

        Returns
        -------
        numpy.array
            The pressure of solidus intersection of each lithology.
        """
        intersect = np.zeros(self.number_lithologies)
        for i in range(self.number_lithologies):
            def f_solve(P):
                return self.lithologies[i].TSolidus(P) - self.adiabat(P, Tp)
            # Check that the lithology actually melts:
            if self.lithologies[i].TSolidus(3.0) is np.inf:
                intersect[i] = np.nan
            else:
                try:
                    intersect[i] = fsolve(f_solve, 3.0)[0]
                except Exception:
                    intersect[i] = np.nan
        return intersect

    def solidus_intersection_isobaric(self, P):
        """
        Finds the pressure at which each lithology's solidus will be intersected,
        assuming the mantle is heated isobarically.

        Parameters
        ----------
        P : loat
            The pressure of interest in GPa

        Returns
        -------
        np.array
            The temperature of solidus intersection of each lithology.
        """
        intersect = np.zeros(self.number_lithologies)
        for i in range(self.number_lithologies):
            intersect[i] = self.lithologies[i].TSolidus(P)
        return intersect

    def adiabat(self, P, Tp):
        """
        Calculates the actual temperature of the solid mantle at a given pressure, given the
        potential temperature.

        Parameters
        ----------
        P :  float or numpy.array
            Pressure in GPa.
        Tp : float or numpy.array
            Potential temperature in degC.

        Returns
        -------
        float or numpy.array
            Temperature of the mantle at the given pressure and Tp.
        """
        bulk_props = self.bulk_properties(P)
        T = ((Tp + 273.15) * np.exp(bulk_props['alpha'] /
             (bulk_props['rho'] * bulk_props['CP']) * P) - 273.15)
        return T

    def F(self, P, T):
        """
        Calculates the melt fraction of each lithology at a given pressure and
        temperature.

        Parameters
        ----------
        P : float
            Pressure in GPa
        T : float
            Temperature in degC

        Returns
        -------
        np.array
            Array containing the melt fraction of each lithology.
        """
        F = np.zeros(self.number_lithologies)
        for i in range(self.number_lithologies):
            F[i] = self.lithologies[i].F(P, T)
        return F

    def dFdP(self, P, T):
        """
        Calculates the value of dFdP for each lithology at the given pressure
        and temperature, using Eq(26) of Phipps Morgan (2001).

        Parameters
        ----------
        P : float
            Pressure in GPa.
        T : float
            Temperature in degC.

        Returns
        -------
        np.array
            Array of dFdP values for each lithology.
        """
        dFdP = np.zeros(self.number_lithologies)

        dTdP = np.zeros(self.number_lithologies)
        dTdF = np.zeros(self.number_lithologies)
        F = np.zeros(self.number_lithologies)

        for i in range(self.number_lithologies):
            dTdP[i] = self.lithologies[i].dTdP(P, T)
            dTdF[i] = self.lithologies[i].dTdF(P, T)
            F[i] = self.lithologies[i].F(P, T)

        lithologies_melting = np.where((F > 0) & (F < 1))[0]

        if np.shape(lithologies_melting)[0] > 0:
            for i in range(np.shape(lithologies_melting)[0]):
                not_key = [True]*self.number_lithologies
                key = lithologies_melting[i]
                not_key[key] = False

                bulk_props = self.bulk_properties(P, T)

                # Equation (26) from PM2001 to find dFdP of first lithology
                top = (bulk_props['CP'] / (T+273.15) * dTdP[key] -
                       bulk_props['alpha']/bulk_props['rho'] +
                       sum(self.proportions[not_key] * self.DeltaS[not_key] *
                           (dTdP[key] - dTdP[not_key]) / dTdF[not_key]))

                bottom = (self.proportions[key] * self.DeltaS[key] +
                          sum(self.proportions[not_key] * self.DeltaS[not_key] *
                              dTdF[key] / dTdF[not_key]) +
                          bulk_props['CP'] / (T+273.15) * dTdF[key])

                dFdP[key] = - top/bottom

        return dFdP

    def adiabatic_gradient(self, P, T):
        """
        Calculates dTdP if melting has gone to completion (or hasn't started) for the bulk mantle.

        Parameters
        ----------
        P : float
            Pressure in GPa.
        T : float
            Temperature in degC.

        Returns
        -------
        float
            The adiabatic gradient in C/GPa
        """

        bulk_props = self.bulk_properties(P, T)

        dTdP = bulk_props['alpha'] * (T+273.15) / bulk_props['rho'] / bulk_props['CP']

        return dTdP

    def dTdP(self, P, T, dFdP=None):
        """
        Calculates dTdP using Eq(28) of Phipps Morgan (2001). Picks the lithology to use by the
        one with the largest increase in melt fraction with decompression (though this choice
        shouldn't matter).

        Parameters
        ----------
        P :    float
            Pressure in GPa.
        T :    float
            Temperature in degC.
        dFdP : np.array or None, default: None
            The value of dFdP at the same T and P. If None, the dFdP method will be called. In
            most melting calculations the value will already have been calculated, so passing the
            value will save computational time.

        Returns
        -------
        float
            The thermal gradient in the melting region at the P and T of interest.
        """
        if dFdP is None:
            dFdP = self.dFdP(P, T)

        melting_lithologies = np.where(dFdP < 0)[0]

        if np.shape(melting_lithologies)[0] > 0:
            key = np.argmin(dFdP[dFdP < 0])
            key = melting_lithologies[key]
            dTdP = self.lithologies[key].dTdP(P, T) + self.lithologies[key].dTdF(P, T) * dFdP[key]
        else:
            dTdP = self.adiabatic_gradient(P, T)

        return dTdP

    def IsobaricMelt_1D(self, Tstart, P, dT=0.1):
        """
        Calculates the amount of melt generated, and the mantle's temperature, after
        an interval of melting occuring due to mantle being instantaneously placed
        above its solidus.

        The intention of this function was to handle high Tp cases where the solidus
        is always exceeded, not to produce physically meaningful results, but to
        allow the tails of Tp distributions to be approximated reasonably when inverting.

        Parameters
        ----------
        Tstart : float
            The temperature (degC) at which to place the solid mantle.
        P :      float
            The pressure at which to perform the calculation.
        dT :     float
            The interval of discretisation of temperature increase from the solidus.

        Returns
        -------
        float
            Temperature of mantle following the melting step.
        """

        solidus_intersection = self.solidus_intersection_isobaric(P)

        # Calculate the entropy lost associated with cooling solid material to the
        # solidus temperature
        solT = np.nanmin(solidus_intersection)
        bulk_props = self.bulk_properties(P, solT)
        DeltaS_cool = - bulk_props['CP'] * np.log((solT + 273.15) / (Tstart + 273.15))

        DeltaS_melt = 0
        T = solT + dT
        while DeltaS_melt < DeltaS_cool and T < Tstart:
            bulk_props = self.bulk_properties(P, T)
            DeltaS_melt = (np.sum(self.F(P, T) * self.proportions * self.DeltaS) +
                           - bulk_props['CP'] * np.log((solT + 273.15) / (T + 273.15)))
            T = T + dT

        return T

    def AdiabaticMelt_1D(self, Tp, Pstart=8.0, Pend=0.01, steps=101, ReportSSS=True):
        """
        Performs simultaneous integration of dFdP and dTdP to obtain the thermal gradient
        through the melting region. F of each lithology is then calculated using the P,T path.
        Integration is performed using a 4th order Runge-Kutta algorithm.

        The T-P path is allowed to overstep the solidus on the step prior to the start of melting.

        Parameters
        ----------
        Tp :        float
            The potential temperature (degC) at which to perform the calculation.
        Pstart :    float, default: 8.0
            The pressure (in GPa) at which to begin upwelling.
        Pend :      float, default: 0.0
            The pressure (in GPa) at which to stop upwelling.
        steps :     int, default: 101
            The number of dP increments to split the melting region into.
        ReportSSS : bool, default: True
            Print to the console if the start is above the solidus of one of the lithologies.
            Either way the code will calculate the melt fraction at this point by conserving
            entropy. Set to False if you don't want to be warned.

        Returns
        -------
        pyMelt.MeltingColumn_1D
            The results are returned in a 1D Melting Column instance, further calculations, e.g.
            crustal thickness may then be performed on this instance, if desired.
        """

        T = np.zeros(steps)
        T[0] = self.adiabat(Pstart, Tp)
        P = np.linspace(Pstart, Pend, steps)
        dP = (Pend - Pstart) / (steps - 1)
        F = np.zeros([steps, self.number_lithologies])

        if T[0] > np.nanmin(self.solidus_intersection_isobaric(Pstart)):
            if ReportSSS is True:
                warn("Super solidus start")
            T[0] = self.IsobaricMelt_1D(T[0], Pstart)

        for i in range(steps):
            if i == 0:
                F[i] = self.F(P[0], T[0])
            else:
                # If melting has not started
                if np.shape(np.where((F[i-1] > 0))[0])[0] == 0:
                    T[i] = self.adiabat(P[i], Tp)
                    F[i] = self.F(P[i], T[i])

                # If melting has gone to completion
                elif np.shape(np.where((F[i-1] > 0) & (F[i-1] < 1))[0])[0] == 0:
                    j1 = self.adiabatic_gradient(P[i-1], T[i-1])
                    j2 = self.adiabatic_gradient(P[i-1] + dP/2, T[i-1] + dP/2*j1)
                    j3 = self.adiabatic_gradient(P[i-1] + dP/2, T[i-1] + dP/2*j2)
                    j4 = self.adiabatic_gradient(P[i], T[i-1] + dP*j3)

                    T[i] = T[i-1] + dP/6*(j1 + 2*j2 + 2*j3 + j4)
                    F[i] = self.F(P[i], T[i])

                # If melting is ongoing
                else:
                    k1 = self.dFdP(P[i-1], T[i-1])
                    j1 = self.dTdP(P[i-1], T[i-1], k1)
                    k2 = self.dFdP(P[i-1] + dP/2, T[i-1] + dP/2*j1)
                    j2 = self.dTdP(P[i-1] + dP/2, T[i-1] + dP/2*j1, k2)
                    k3 = self.dFdP(P[i-1] + dP/2, T[i-1] + dP/2*j2)
                    j3 = self.dTdP(P[i-1] + dP/2, T[i-1] + dP/2*j2, k3)
                    k4 = self.dFdP(P[i], T[i-1] + dP*j3)
                    j4 = self.dTdP(P[i], T[i-1] + dP*j3, k4)

                    T[i] = T[i-1] + dP/6 * (j1 + 2*j2 + 2*j3 + j4)
                    F[i] = self.F(P[i], T[i])

        results = pd.DataFrame(F, columns=self.names)
        results['P'] = P
        results['T'] = T

        return MeltingColumn_1D(results, self, Tp)



"""
===========
Lithologies
===========

The lithologies module provides lithology objects representing different bulk compositions, with
models published by a number of authors.

"""

from pyMelt.lithologies import matthews
from pyMelt.lithologies import katz
from pyMelt.lithologies import pertermann
from pyMelt.lithologies import shorttle
from pyMelt.lithologies import mckenzie

__all__ = ['matthews', 'katz', 'pertermann', 'shorttle', 'mckenzie']



"""
===============
Melting Columns
===============

the meltingcolumn_classes module provides the melting column classes. At the moment it consists of
a single melting column- a simple 1D melting column.

"""

import numpy as np
from pyMelt.core import InputError
import matplotlib.pyplot as plt


class MeltingColumn_1D():
    """
    Class for storing the results of a 1D multi-lithology melting model.

    Parameters
    ----------
    calculation_results : pandas.DataFrame
        Dataframe with columns 'P' for Pressure in GPa, 'T' for Temperature in
        degrees C, Remaining columns for melt fraction from each lithology.
    mantle : pyMelt.Mantle
        The mantle class used to generate the melting column.
    Tp : float
        The potential temperature used to generate the melting column, if applicable.

    Attributes
    ----------
    P : pandas.Series
        The pressures of the melting column in GPa.
    T : pandas.Series
        The temperatures of the melting column in degC at each pressure.
    F : pandas.DataFrame
        The melt fractions of each lithology at each pressure.
    mantle : pyMelt.Mantle
        The mantle class used to generate the melting model
    Tp : float
        The potential temperature (in degC) used to generate the melts.
    F_total : numpy.array
        The total melt fraction at each pressure.
    dtcdP : numpy.array or None
        The results from Eq6 for the total melt fraction following integration. Available following
        integration.
    tc_int : numpy.array or None
        Integrated crustal thickness as a function of pressure (up to 0 GPa). Available following
        integration.
    tc_P_int : numpy.array or None
        The pressure exerted by the integrated crustal thickness as a function of pressure (up to
        0 GPa). Available following integration.
    tc : float or None
        The integrated crustal thickness at the point where the pressure it exerts is equal to the
        calculation pressure. Available following integration.
    P_base_of_crust : float or None
        The pressure at the base of the crust, at the point where the pressure the generated crust
        exerts is equal to the calculation pressure. Available following integration.
    tc_lithology_contributions_int : pandas.DataFrame or None
        The integrated proportion of generated crust derived from each lithology as a function of
        pressure. Available following integration.
    tc_lithology_contributions : pandas.Series or None
        The integrated proportion of generated crust derived from each lithology at the pressure
        where P(calculation) = P(exerted by generated crust). Available following integration.
    """

    def __init__(self, calculation_results, mantle, Tp=None):
        self.P = calculation_results.P
        self.T = calculation_results['T']

        cols = calculation_results.columns.tolist()
        cols.remove('P')
        cols.remove('T')
        self.F = calculation_results[cols]

        self.mantle = mantle
        self.Tp = Tp
        self.F_total = np.zeros(np.shape(self.F)[0])

        for i in range(self.mantle.number_lithologies):
            self.F_total = self.F_total + self.mantle.proportions[i]*self.F[self.mantle.names[i]]

        self.dtcdP = None
        self.tc_int = None
        self.tc_P_int = None
        self.tc = None
        self.P_base_of_crust = None
        self.tc_lithology_contributions_int = None
        self.tc_lithology_contributions = None

    def plot(self):
        """
        Generates a plot showing the thermal gradient and melt fractions ofeach lithology.

        Returns
        -------
        (matplotlib.figure, matplotlib.axes)
            The generated figure and axes objects.
        """
        f, a = plt.subplots(1, 2, sharey='row', dpi=100)

        lith = self.F.columns

        for i in range(np.shape(lith)[0]):
            a[1].plot(self.F.iloc[:, i], self.P, label=lith[i])

        a[0].plot(self.T, self.P, label='Thermal Gradient', c='k')
        a[1].plot(self.F_total, self.P, label='Total', c='k', ls='--')

        P = np.linspace(np.min(self.P), np.max(self.P), 101)
        for i in range(self.mantle.number_lithologies):
            T = self.mantle.lithologies[i].TSolidus(P)
            a[0].plot(T, P, label=self.mantle.names[i]+' solidus')

        if self.Tp is not None:
            a[0].text(0.95, 0.95, 'T$_p$ = {:.0f} °C'.format(self.Tp),
                      transform=a[0].transAxes, va='top', ha='right')

        a[0].invert_yaxis()

        a[0].set_ylabel('Pressure (GPa)')
        a[0].set_xlabel('Temperature (°C)')
        a[1].set_xlabel('Melt Fraction')

        a[0].legend(loc='lower left')
        a[1].legend()

        a[0].tick_params('x', labeltop=True, labelbottom=False)
        a[0].xaxis.set_label_position('top')
        a[1].tick_params('x', labeltop=True, labelbottom=False)
        a[1].xaxis.set_label_position('top')

        return f, a

    def integrate_tri(self, P_base_existingLith=0.0, extract_melt=False):
        """
        Perform an integration over the melting region, assuming it is triangular and passively
        upwelling.

        Parameters
        ----------
        P_base_existingLith : float, default: 0.0
            The pressure at the base of any pre-existing lithosphere. The calculated thickness will
            be added to this value. Set to 0.0 for mid-ocean ridges, set to non-zero for
            continental rifting.

        extract_melt : bool, default: False
            If set to True, the melts will be extracted from the system. If False (default) the
            melt produced is added to the top of the melting column, and the calculation will stop
            once the pressure exerted from the newly made crust is equal to the pressure in the
            calculation step.

        Returns
        -------
        float
            The crustal thickness where the pressure exerted from the produced crust equals the
            pressure of melting.

        """
        rho = self.mantle.bulk_properties()['rho']
        g = 9.81
        tc = 1.0 / (rho * g * 1e3) * self.F_total / (1.0 - self.F_total)
        tc_lith = (1.0 / (rho * g * 1e3) * self.F * self.mantle.proportions
                   / (1.0 - np.tile(self.F_total, [np.shape(self.F)[1], 1]).T))
        tc_int = np.zeros(np.shape(self.P)[0])
        tc_lith_int = np.zeros(np.shape(tc_lith))
        tc_intP = np.zeros(np.shape(self.P)[0])
        tc_found = False
        P_basecrust = False
        tc_lith_found = False

        for i in range(np.shape(self.P)[0]):
            if i != 0:
                tc_int[i] = tc_int[i-1] + tc[i] * np.abs(self.P[i] - self.P[i-1])
                tc_lith_int[i] = (tc_lith_int[i-1]
                                  + tc_lith.iloc[i] * np.abs(self.P[i] - self.P[i-1]))
                tc_intP[i] = tc_int[i] * rho * g * 1e3
                if(extract_melt is False and tc_intP[i] + P_base_existingLith > self.P[i]
                   and tc_found is False):
                    tc_found = tc_int[i]
                    P_basecrust = self.P[i]
                    tc_lith_found = tc_lith_int[i]
                elif (extract_melt is True
                      and (i == np.shape(self.P)[0] - 1 or P_base_existingLith > self.P[i])
                      and tc_found is False):
                    tc_found = tc_int[i]
                    P_basecrust = self.P[i]
                    tc_lith_found = tc_lith_int[i]

        self.dtcdP = tc
        self.tc_int = tc_int * 1e6
        self.tc_P_int = tc_intP
        self.tc = tc_found * 1e6
        self.P_base_of_crust = P_basecrust
        self.tc_lithology_contributions_int = tc_lith / tc_lith.sum()
        self.tc_lithology_contributions = tc_lith_found / sum(tc_lith_found)

        return tc_found * 1e6

    def MeltCrystallisationT(self, ShallowMeltP=None, MeltStorageP=None, liqdTdP=39.16):
        """
        Identifies the crystallisation temperature of the deepest and shallowest melts,
        according to the technique used by Matthews et al. (2016).

        Parameters
        ----------
        ShallowMeltP : float or None, default: None
            The pressure (in GPa) at which the shallowest melt should be extracted. If set
            to None (as is default) this will be taken as the base of the crust. If integration
            has not been performed, this will result in an error.
        MeltStorageP : float, default:  None
            The pressure at which crystallisation is happening. If set to False (as is default),
            the base of the crust will be used. If triangular integration has not been done,
            this will result in an error.
        liqdTdP : float, default: 39.16
            The clapeyron slope of the liquidus (K GPa-1), the default value is 39.16,
            from equation (15) of Putirka  (2008).

        Returns
        -------
        float
            The minimum crystallisation temperature (degC)
        float
            The maximum crystallisation temperature (degC)
        """

        # Set crystallisation Pressure
        if MeltStorageP is None:
            if self.P_base_of_crust is None:
                raise InputError("You must perform an integration before calculating Tcrys.")
            MeltStorageP = self.P_base_of_crust
        if ShallowMeltP is None:
            ShallowMeltP = self.P_base_of_crust

        self.T_crystallisation = self.T - (self.P - MeltStorageP) * liqdTdP

        self.DeepMeltTcrys = {}
        for l in self.mantle.names:
            if (self.F[l] > 0).any():
                FirstMeltRow = self.F[l][self.F[l] > 0].idxmin()
                self.DeepMeltTcrys[l] = self.T_crystallisation.iloc[FirstMeltRow]
            else:
                self.DeepMeltTcrys[l] = np.nan
        self.TcrysMax = np.nanmax(list(self.DeepMeltTcrys.values()))

        LastMeltRow = self.P[self.P > ShallowMeltP].idxmin()
        self.TcrysMin = self.T_crystallisation.iloc[LastMeltRow]

        return self.TcrysMin, self.TcrysMax



"""
=================
Lithology classes
=================

The lithology classes module provides the `lithology` class for implementing melting models, and
the `hydrous_lithology` class for converting an anhydrous lithology to a hydrous lithology.
"""
from scipy.misc import derivative
from scipy.optimize import root_scalar
import numpy as _np
from pyMelt.core import ConvergenceError

# Default constant values taken from Katz et al., 2003:
default_properties = {'CP':     1000.0,  # Heat capacity in J Kg-1 K-1
                      'alphas':   40.0,  # Thermal expansivity of the solid (K-1).
                      'alphaf':   68.0,  # Thermal expansivity of the melt (K-1).
                      'rhos':      3.3,  # Density of the solid (g cm-3).
                      'rhof':      2.9,  # Density of the melt (g cm-3).
                      'DeltaS':  300.0,  # Entropy of fusion. (J kg-1 K-1).
                      }


class lithology(object):
    """
    Lithology base class. This class contains all the parameters and methods required to calculate
    the melting behaviour of a single mantle component.

    Attributes
    ----------
    DeltaS:     float
        Entropy of fusion. (J kg-1 K-1). Default is 300.0.
    CP:     float
        Heat capacity (J Kg-1 K-1). Default is 1000.0.
    alphas:     float
        Thermal expansivity of the solid (K-1). Default is 40.0.
    alphaf:     float
        Thermal expansivity of the melt (K-1). Default is 68.0.
    rhos:   float
        Density of the solid (g cm-3). Default is 3.3.
    rhof:   float
        Density of the melt (g cm-3). Default is 2.9.

    """

    def __init__(self,
                 CP=default_properties['CP'],
                 alphas=default_properties['alphas'],
                 alphaf=default_properties['alphaf'],
                 rhos=default_properties['rhos'],
                 rhof=default_properties['rhof'],
                 DeltaS=default_properties['DeltaS'],
                 parameters={}):
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.DeltaS = DeltaS
        self.parameters = parameters

    def TSolidus(self, P):
        """
        Returns the temperature of the solidus at any given pressure.

        Parameters
        ----------
        P:  float
            Pressure (GPa).

        Returns
        -------
        None
            The default lithology never melts.
        """

        return None

    def TLiquidus(self, P):
        """
        Returns the temperature of the liquidus at any given pressure.

        Parameters
        ----------
        P:  float
            Pressure (GPa).

        Returns
        -------
        None
            The default lithology never melts.
        """
        return None

    def F(self, P, T):
        """
        Returns the melt fraction at any given pressure and temperature.

        Parameters
        ----------
        P:  float
            Pressure (GPa).
        T:  float
            Temperature (degC).

        Returns
        -------
        float
            Melt fraction.
        """

        return 0.0

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P) numerically. For faster and more accurate results redefine
        this method.

        Parameters
        ----------
        P:  float
            Pressure (GPa)
        T:  float
            Temperature (degC)

        Returns
        -------
        float
            dT/dF(const. P) (K).
        """

        def _to_diff(T, P, kwargs={}):
            return self.F(P, T, **kwargs)

        return 1.0/(derivative(_to_diff, T, dx=0.1, args=(P, kwargs)))

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F) numerically. For faster and more accurate results redefine
        this method.

        Parameters
        ----------
        P:  float
            Pressure (GPa).
        T:  float
            Temperature (degC).

        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """
        dTdF = self.dTdF(P, T, **kwargs)

        def _to_diff(P, T, kwargs={}):
            return self.F(P, T, **kwargs)

        dFdP = derivative(_to_diff, P, args=(T, kwargs))

        return dTdF*dFdP


class hydrous_lithology(object):
    """
    The hydrous lithology class modifies the melting expressions in a given lithology class so
    that water-present melting can be modelled.

    Parameters
    ----------
    lithology : pyMelt.lithology_classes.lithology instance
        The anhydrous lithology for which hydrous melting should be modelled.
    H2O : float
        The water content of the lithology in wt%.
    D : float, default: 0.01
        Partition coefficient for water partitioning between melt and the (bulk) residue.
    K : float, default: 43.0
        Parameter controlling hydrous solidus position, from Katz et al. (2003). See notes.
    gamma : float, default: 0.75
        Parameter controlling the hydrous solidus position, from Katz et al. (2003). Must take a
        value between 0 and 1. See notes.
    chi1 : float, default: 12.0
        Parameter controlling the H2O content of melt at H2O-satuation, in wt% GPa^(-l).
    chi2 : float, default: 1.0
        Parameter controlling the H2O content of melt at H2O-satuation, in wt% GPa^(-1).
    l : float, default: 0.6
        Parameter controlling the H2O content of melt at H2O-satuation.
    continuous : bool, default: False
        Controls whether water extracted to melt is kept or removed from the system. If False,
        melting will be done assuming batch closed-system melting. If True, the water contents
        are modelled according to continuous melting, i.e., near-fractional melting, with the
        porosity controlled by the phi argument.
    phi : float, default: 0.005
        The porosity to use during continuous melting, if applicable. The default value is 0.5%.
    threshold_F : float or None, default: None
        If set, when the melt fraction goes above the chosen threshold, the hydrous extension will
        be skipped. Doing so may help speed up the calculations, but a threshold melt fraction
        should be chosen with care.

    Notes
    -----
    The parameters controlling the position of the hydrous solidus are defined by eqn. 16 from
    Katz et al. (2003):

    .. math::

        \Delta T(X_{H2O}) = K X^\gamma_{H2O}

    Where :math:`X_{H2O}` is the water content in the melt.

    The water content at saturation is controlled by eqn. 17 of Katz et al. (2003):

    .. math::

        X^{sat}_{H2O} = \chi_1 P^\lambda + \chi_2 P


    """

    def __init__(self, lithology, H2O, D=0.01, K=43.0, gamma=0.75, chi1=12.0, chi2=1.0,
                 l=0.6, continuous=False, phi=0.005, threshold_F=None):
        self.lithology = lithology

        for m in dir(lithology):
            if '__' not in m and m not in ['TSolidus', 'TLiquidus', 'F', 'dTdF', 'dTdP']:
                f = lithology.__getattribute__(m)
                if callable(f):
                    self.__setattr__(m, f.__func__.__get__(self, self.__class__))
                else:
                    self.__setattr__(m, f)

        self.F_function = self.lithology.F.__func__.__get__(self, self.__class__)
        self.H2O = H2O
        self.K = K
        self.gamma = gamma
        self.D = D
        self.chi1 = chi1
        self.chi2 = chi2
        self.l = l
        self.continuous = continuous
        self.phi = phi
        self.threshold_F = threshold_F

    def H2O_saturation(self, P):
        """
        The H2O content of the melt at H2O saturation. This is Eqn 17 of Katz et al. (2003).

        Parameters
        ----------
        P : float
            Pressure in GPa

        Returns
        -------
        float
            The H2O content of the melt at H2O-saturation, in wt%.
        """

        return self.chi1 * P ** self.l + self.chi2 * P

    def melt_H2O(self, F, P=None):
        """
        Returns the H2O content of the melt at a particular melt fraction, assuming batch
        melting, and H2O saturation controlled by the H2O_saturation method. If H2O is
        saturated, it will remain in the bulk system.

        Parameters
        ----------
        F : float
            The melt fraction. Must be between 0 and 1.
        P : float or None, default: None
            The pressure in GPa. Used for checking for H2O saturation. If set to None this
            check will be bypassed.

        Returns
        -------
        float
            The H2O content of the melt, in wt%.
        """

        if self.continuous is False:
            Cl = self.H2O/((1.0 - F)*self.D + F)
        else:
            Cl = (self.H2O / ((1-self.phi)*self.D + self.phi)
                  * (1-F)**((1-self.phi)*(1-self.D)/((1-self.phi)*self.D)))

        # Test for H2O saturation:
        if P is not None:
            H2Osat = self.H2O_saturation(P)
            if H2Osat < Cl:
                Cl = H2Osat

        return Cl

    def TSolidus(self, P, F=0.0):
        """
        Returns the temperature of the solidus at any given pressure. Since the solidus is a
        function of the water content of the bulk, the effective solidus position will shift
        during melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        F : float, default: 0.0
            Melt fraction, between 0 and 1.

        Returns
        -------
        Tsol:   float
            Solidus temperature (degC).
        """

        TSolidus = self.lithology.TSolidus(P) - self.K * self.melt_H2O(F, P)**self.gamma

        return TSolidus

    def TLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the liquidus at any given pressure.
        Parameters
        ----------
        P:  float
            Pressure (GPa).
        Returns
        -------
        Tliq:   float
            Liquidus temperature (degC).
        """
        return self.lithology.TLiquidus(P)

    def F(self, P, T, **kwargs):
        """
        Returns the melt fraction of the hydrated lithology, by iteratively finding the value of F
        which provides the amount of water to the melt such that the same value of F is returned
        by the calculation. This is equivalent to Eqn 19 of Katz et al. (2003), but its form will
        depend on the lithology being used.

        The numerical problem is solved using the SciPy method root_scalar from the optimize
        module. The method used is brentq, so that the problem may be constrained between a
        melt fraction of 0 and 1.

        Parameters
        ----------
        P : float
            Pressure in GPa
        T : float
            Temperature in degC.

        Returns
        -------
        float
            The melt fraction.

        """
        if self.threshold_F is not None:
            anhydrous_F = self.lithology.F(P, T, **kwargs)
            if anhydrous_F > self.threshold_F:
                return anhydrous_F

        F = root_scalar(self._f_to_solve, bracket=[0, 1], args=(P, T))

        if F.flag != 'converged':
            raise convergenceError('The melt fraction calculation did not converge.')
            return _np.nan
        else:
            return F.root

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P) for the hydrated lithology numerically.

        Parameters
        ----------
        P:  float
            Pressure (GPa)
        T:  float
            Temperature (degC)

        Returns
        -------
        float
            dT/dF(const. P) (K).
        """

        return 1.0/(derivative(self._to_diff_dTdF, T, dx=0.001, args=(P, kwargs)))

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F) for the hydrated lithology numerically.

        The method uses the root_scalar method of the scipy.optimize module to find the
        curve of T-P for which F=const. The scipy derivative method is then used to numerically
        differentiate this curve.

        Parameters
        ----------
        P:  float
            Pressure (GPa).
        T:  float
            Temperature (degC).

        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """
        if self.F(P, T, **kwargs) == 0:
            dTdP = self.alphas/self.rhos/self.CP
        elif self.F(P, T, **kwargs) == 1:
            dTdP = self.alphas/self.rhos/self.CP
        else:
            F = self.F(P, T, **kwargs)

            dTdP = derivative(self._to_diff_dTdP, P, dx=0.001, args=(F, T, kwargs))

        return dTdP

    # The following methods support the numerical differentiation and root-finding methods used
    # above. The reason they are defined as methods, and do not sit within their "parent" methods
    # is to assist with module testing and troubleshooting.

    # This method rearranges the arguments so that F can be numerically differentiated.
    def _to_diff_dTdF(self, T, P, kwargs={}):
        return self.F(P, T, **kwargs)

    # This method is used to find the P-T curve at which F remains constant, for the calculation of
    # dT/dP (at const. F).
    def _hold_constant_F(self, t, P, F, kwargs={}):
        return self.F(P, t, **kwargs) - F

    # This finds the temperature at a given pressure for which the melt fraction is eqal to the
    # value specified. Used for calculating dT/dP (at const. F).
    def _to_diff_dTdP(self, P, F, T, kwargs={}):
        t = root_scalar(self._hold_constant_F, x0=T, x1=T+10, args=(P, F, kwargs)).root
        return t

    # This provides the misfit function when solving for F (Eqn. 19 of Katz et al., 2003)
    def _f_to_solve(self, x, P, T):
        misfit = 0
        fcalc = self.F_function(P, T, F=x)
        misfit = fcalc - x
        return misfit



"""
==================
McKenzie and Bickle (1988)
==================

Implementation of the anhydrous melting model presented by McKenzie and Bickle (1988).

"""

from pyMelt.lithology_classes import lithology as _lithology
from pyMelt.lithology_classes import default_properties as _default_properties
import numpy as np
from scipy.optimize import fsolve
from scipy.special import expit


class lherzolite(_lithology):
    """
    Implementation of the McKenzie and Bickle (1988) garnet peridotite melting model.
    As this parameterisation provides pressure as a function of solidus temperature,
    scipt.optimize.fsolve and scipy.special.expit are required to numerically find
    solidus temperature as a function of pressure.
    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class, with
    values:
    - A1:  Parameter used in solidus definition.
    - A2:  Parameter used in solidus definition.
    - A3:  Parameter used in solidus definition.
    - A4:  Parameter used in solidus definition.
    - B1:  Parameter used in liquidus definition.
    - B2:  Parameter used in liquidus definition.
    - B3:  Parameter used in liquidus definition.
    - B4:  Parameter used in liquidus definition.
    - a0:  Parameter used in calculating melt fraction.
    - a1:  Parameter used in calculating melt fraction.
    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.
    Parameters
    ----------
    CP :         float, default: pyMelt.lithology_class.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.lithology_class.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.lithology_class.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.lithology_class.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.lithology_class.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.lithology_class.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from McKenzie and Bickle (1988)
        The model parameters described above
    """
    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'A1':   1100.0,
                             'A2':    136.0,
                             'A3': 4.968e-4,
                             'A4':   1.2e-2,
                             'B1':   1736.0,
                             'B2':    4.343,
                             'B3':    180.0,
                             'B4':   2.2169,
                             'a0':   0.4256,
                             'a1':    2.988
                             }
                 ):
        self.DeltaS = DeltaS
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.parameters = parameters

    def TSolidus(self, P):
        """
        Calculates the solidus temperature, at a given pressure, using Equation 18 of
        McKenzie and Bickle (1988). Requires scipy.optimize.fsolve to solve for pressure.
        As fsolve can only solve for a single pressure TSolidus first detects whether P is
        given as a list, tuple, numpy.ndarray (e.g. numpy.linspace), or scalar prior to
        calculating the solidus pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        Returns
        -------
        float
            Solidus temperature (degC).
        """
        TSolidus_initial_guess = 1300.0

        def func(TSolidus):
            _func = ((TSolidus - self.parameters['A1']) / self.parameters['A2']
                     + self.parameters['A3'] * np.exp(self.parameters['A4']
                     * (TSolidus - self.parameters['A1']))) - Pressure
            return _func

        if isinstance(P, (list, tuple, np.ndarray)) is True:
            TSolidus_solution = []
            for i in range(len(P)):
                Pressure = P[i]
                TSolidus_solution.append(fsolve(func, TSolidus_initial_guess, maxfev=50000)[0])
        else:
            Pressure = P
            TSolidus_solution = fsolve(func, TSolidus_initial_guess, maxfev=50000)[0]
        return TSolidus_solution

    def TLiquidus(self, P):
        """
        Calculates the liquidus temperature, at a given pressure, using Equation 19 of
        McKenzie and Bickle (1988).

        Parameters
        ----------
        P : float
            Pressure (GPa)
        Returns
        -------
        float
            Solidus temperature (degC).
        """
        TLiquidus = (self.parameters['B1'] + self.parameters['B2'] * P + self.parameters['B3']
                     * np.arctan(P / self.parameters['B4']))
        return TLiquidus

    def _dTdPSolidus(self, P):
        """
        Returns the solidus temperature gradient at any given pressure. Requires
        scipy.optimize.fsolve to solve for pressure, and scipy.special.expit to
        avoid exponential overflows that can occur with numpy.exp.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        Returns
        -------
        float
            Solidus temperaure gradient (degC/GPa)
        """
        TSolidus = self.TSolidus(P)
        dTdPSolidus = (1 / (1 / self.parameters['A2'] + self.parameters['A3']
                       * self.parameters['A4'] * expit(self.parameters['A4']
                       * (TSolidus - self.parameters['A1']))))
        return dTdPSolidus

    def _dTdPLiquidus(self, P):
        """
        Returns the liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        Returns
        -------
        float
            Solidus temperaure gradient (degC/GPa)
        """
        dTdPLiquidus = (self.parameters['B2'] + self.parameters['B3'] /
                        (1 + (P / self.parameters['B4'])**2))
        return dTdPLiquidus

    def _RescaledT(self, T, P):
        """
        Calculates the rescaled temperature defined by Equation 20 of
        McKenzie and Bickle (1988).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (GPa).
        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TSolidus = self.TSolidus(P)
        TLiquidus = self.TLiquidus(P)
        RescaledT = (T - (TSolidus + TLiquidus) / 2) / (TLiquidus - TSolidus)
        return RescaledT

    def F(self, P, T):
        """
        Wrapper for the melt fraction functions, defined by Equations 21 and 22
        of McKenzie and Bickle (1988). If T and P are below the solidus, returns 0,
        if they are above the liquidus, returns 1. Otherwise determines F.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).
        Returns
        -------
        float
            Melt fraction.
        """
        TSolidus = self.TSolidus(P)
        TLiquidus = self.TLiquidus(P)
        RescaledT = self._RescaledT(T, P)
        if T > TLiquidus:
            F = 1.0
        elif T < TSolidus:
            F = 0.0
        else:
            F = (RescaledT + (RescaledT**2 - 0.25) * (self.parameters['a0']
                 + self.parameters['a1'] * RescaledT) + 0.5)
        return F

    def dTdF(self, P, T):
        """
        Calculates dT/dF(const. P). First calculates the melt fraction. If F is
        zero, returns _np.inf. If F is 1, returns _np.inf. Otherwise uses the
        appropriate expression for melt fraction.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)
        Returns
        -------
        float
            dT/dF(const. P) (K).
        """
        TSolidus = self.TSolidus(P)
        TLiquidus = self.TLiquidus(P)
        RescaledT = self._RescaledT(T, P)
        if T < TSolidus:
            dTdF = np.inf
        elif T > TLiquidus:
            dTdF = np.inf
        else:
            dFdT = (((1 - 0.25 * self.parameters['a1']) + 3 * self.parameters['a0'] * RescaledT
                     + 3 * self.parameters['a1'] * RescaledT**2) / (TLiquidus - TSolidus))
            dTdF = 1 / dFdT
        return dTdF

    def dTdP(self, P, T):
        """
        Calculates dT/dP(const. F). First calculates F, then chooses the
        appropriate expression for melt fraction.

        Parameters
        ----------
        P:  float
            Pressure (GPa).
        T:  float
            Temperature (degC).
        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """
        dTdPSolidus = self._dTdPSolidus(P)
        dTdPLiquidus = self._dTdPLiquidus(P)
        RescaledT = self._RescaledT(T, P)
        dTdP = RescaledT * (dTdPLiquidus - dTdPSolidus) + 0.5 * (dTdPSolidus + dTdPLiquidus)
        return dTdP



"""
======================
Matthews et al. (2021)
======================

Implentation of the melting models developed by Matthews et al. (2021).
"""

from pyMelt.lithology_classes import lithology as _lithology
from pyMelt.lithology_classes import default_properties as _default_properties


import numpy as _np


class kg1(_lithology):
    """
    Implementation of the KG1 melting model from Matthews et al. (2021).

    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class.

    - Mcpx:   Mass fraction of cpx in the source. Controls the transition to
      low-productivity harzburgite-type melting.
    - A1:     Parameter used to define solidus.
    - A2:     Parameter used to define solidus.
    - A3:     Parameter used to define solidus.
    - A4:     Parameter used to define solidus.
    - B1:     Parameter used to define liquidus.
    - B2:     Parameter used to define liquidus.
    - B3:     Parameter used to define liquidus.
    - B4:     Parameter used to define liquidus.
    - C:      Parameter used to define lherzolite-liquidus.
    - beta1:  Parameter used to calculate melt fraction during cpx-present melting.
    - beta2:  Parameter used to calculate melt fraction during cpx-absent melting.
    - r1:     Parameter used to define cpx reaction coefficient.
    - r2:     Parameter used to define cpx reaction coefficient.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Matthews et al. (2021)
        The model parameters described above

    """

    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'A1':    450.000,
                             'A2':      2.098,
                             'A3':     17.000,
                             'A4':    623.828,
                             'B1':    174.566,
                             'B2':    336.833,
                             'B3':     66.762,
                             'B4':    503.101,
                             'C':       0.506,
                             'beta1':   1.382,
                             'beta2':   1.800,
                             'r1':      0.342,
                             'r2':      0.191,
                             'Mcpx':    0.342
                             }
                 ):
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.DeltaS = DeltaS
        self.parameters = parameters

    def TSolidus(self, P, **kwargs):
        """
        Returns the temperature of the solidus at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperature (degC).
        """
        TSolidus = (self.parameters['A1']*_np.log(P + self.parameters['A2'])
                    + self.parameters['A3']*P
                    + self.parameters['A4'])
        return TSolidus

    def TLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the liquidus at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperature (degC).
        """
        TLiquidus = (self.parameters['B1']*_np.log(P + self.parameters['B2'])
                     + self.parameters['B3']*P + self.parameters['B4'])
        return TLiquidus

    def F(self, P, T, **kwargs):
        """
        Wrapper for the melt fraction functions. If T and P are below the solidus,
        returns 0, if they are above the liquidus, returns 1. If below the temperature
        of cpx-exhaustion, calls the Fcpx function, otherwise calls the Fopx function.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            Melt fraction.
        """
        if T > self.TLiquidus(P, **kwargs):
            F = 1.0
        elif T < self.TSolidus(P, **kwargs):
            F = 0.0
        elif T < self._TcpxOut(P, **kwargs):
            F = self._Fcpx(T, P, **kwargs)
        else:
            F = self._Fopx(T, P, **kwargs)

        return F

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P). First calculates the melt fraction. If F is
        zero, returns _np.inf. If F is 1, returns _np.inf. Otherwise uses the
        appropriate expressions for cpx present or absent melting.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)

        Returns
        -------
        float
            dT/dF(const. P) (K).
        """
        F = self.F(P, T, **kwargs)
        if F == 0:
            # If no melt fraction the derivative is zero. Prevents division by zero.
            dTdF = _np.inf
        elif F < self._FcpxOut(P, **kwargs):
            dTdF = (((1/self.parameters['beta1']))
                    * (self._TLherzLiquidus(P, **kwargs) - self.TSolidus(P, **kwargs))
                    * (F**((1-self.parameters['beta1'])/self.parameters['beta1'])))
        elif F < 1.0:
            dTdF = (((1/self.parameters['beta2']))
                    * (self.TLiquidus(P, **kwargs) - self._TcpxOut(P, **kwargs))
                    * (F**((1-self.parameters['beta2'])/self.parameters['beta2'])))
        else:
            dTdF = _np.inf
        return dTdF

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F). First calculates F, then chooses the
        appropriate expression for cpx present or absent melting.

        Parameters
        ----------
        P:  float
            Pressure (GPa).
        T:  float
            Temperature (degC).

        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """
        F = self.F(P, T, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        dTdPLiquidus = self._dTdPLiquidus(P, **kwargs)
        dTdPLherzLiquidus = self._dTdPLherzLiquidus(P, **kwargs)
        TcpxOut = self._dTdPcpxOut(P, **kwargs)
        dTdPcpxOut = self._dTdPcpxOut(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dFdPcpxOut = self._dFdPcpxOut(P, **kwargs)

        if F == 0:
            dTdP = self.alphas/self.rhos/self.CP
        elif F < self._FcpxOut(P, **kwargs):
            dTdP = (((F**(1/self.parameters['beta1'])) * (dTdPLherzLiquidus-dTdPSolidus))
                    + dTdPSolidus)
        elif F < 1.0:
            Trel = (T - TcpxOut) / (TLiquidus - TcpxOut)
            dTdP = ((TLiquidus - TcpxOut)/(1-FcpxOut)
                    * (1/self.parameters['beta2'])*Trel**(1-self.parameters['beta2'])
                    * dFdPcpxOut * (Trel**self.parameters['beta2']-1)
                    + dTdPcpxOut + Trel*(dTdPLiquidus - dTdPcpxOut))
        else:
            dTdP = self.alphaf/self.rhof/self.CP
        return dTdP

    def _dTdPSolidus(self, P, **kwargs):
        """
        Returns the solidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperaure gradient (degC/GPa)
        """
        dTdPSolidus = (self.parameters['A1']/(P + self.parameters['A2'])) + self.parameters['A3']
        return dTdPSolidus

    def _dTdPLiquidus(self, P, **kwargs):
        """
        Returns the liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperaure gradient (degC/GPa)
            """
        dTdPLiquidus = ((self.parameters['B1']/(P + self.parameters['B2']))
                        + self.parameters['B3'])
        return dTdPLiquidus

    def _TLherzLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the lherzolite liquidus at any given pressure.
        This is the temperature at which the rock would be completely
        molten if cpx was remained present for the entirety of melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Lherzolite liquidus temperature (degC).
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        TLherzLiquidus = self.parameters['C']*TSolidus + (1 - self.parameters['C'])*TLiquidus
        return TLherzLiquidus

    def _dTdPLherzLiquidus(self, P, **kwargs):
        """
        Returns the lherzolite liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Lherzolite liquidus temperaure gradient (degC/GPa)
        """
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        dTdPLiquidus = self._dTdPLiquidus(P, **kwargs)
        dTdPLherzoliteLiquidus = (self.parameters['C']*dTdPSolidus
                                  + (1 - self.parameters['C'])*dTdPLiquidus)
        return dTdPLherzoliteLiquidus

    def _RescaledTcpx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-present melting. (Eq 3).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        RescaledTemperaturecpx = ((T - TSolidus)/(TLherzLiquidus-TSolidus))
        return RescaledTemperaturecpx

    def _Fcpx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-present melting at the given P and T. Eq(2).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-present melting.
        """
        RescaledT = self._RescaledTcpx(T, P, **kwargs)
        Fcpx = RescaledT**self.parameters['beta1']
        return Fcpx

    def _RxnCoef(self, P, **kwargs):
        """
        Reaction coefficient for cpx during melting at the specified pressure. Eq(7).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Reaction Coefficient.
        """
        RxnCoef = self.parameters['r1'] + self.parameters['r2']*P
        return RxnCoef

    def _FcpxOut(self, P, **kwargs):
        """
        Calculates the melt fraction required to exhaust cpx from the residue
        at the given pressure. Eq(6).

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Melt fraction at which cpx is exhausted.
        """
        FcpxOut = self.parameters['Mcpx'] / self._RxnCoef(P, **kwargs)
        return FcpxOut

    def _dFdPcpxOut(self, P, **kwargs):
        """
        Calculates the first derivative of FcpxOut.

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            First derivative of FcpxOut.
        """
        RxnCoef = self._RxnCoef(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dFdPcpxOut = ((-1)*FcpxOut*self.parameters['r2'])/RxnCoef
        return dFdPcpxOut

    def _TcpxOut(self, P, **kwargs):
        """
        Calculates the temperature at which cpx will be exhausted during melting
        at the given pressure. Eq(9).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Temperature of cpx-exhaustion.
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        TcpxOut = (((FcpxOut**(1/self.parameters['beta1']))) * (TLherzLiquidus - TSolidus)
                   + TSolidus)
        return TcpxOut

    def _dTdPcpxOut(self, P, **kwargs):
        """
        Calculates the temperature gradient of the cpx-exhaustion surface.

        Parameters
        ----------

        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Temperature gradient of cpx-exhaustion.
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        dTdPLherzLiquidus = self._dTdPLherzLiquidus(P, **kwargs)
        dFdPcpxOut = self._dFdPcpxOut(P, **kwargs)
        A = ((dFdPcpxOut*(1/self.parameters['beta1'])
             * ((FcpxOut**((1/self.parameters['beta1'])-1)))) *
             (TLherzLiquidus-TSolidus))
        B = ((FcpxOut**(1/self.parameters['beta1']))*(dTdPLherzLiquidus-dTdPSolidus))+dTdPSolidus
        dTdPcpxOut = A + B
        return dTdPcpxOut

    def _RescaledTopx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-absent melting.

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TcpxOut = self._TcpxOut(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        RescaledTopx = ((T-TcpxOut)/(TLiquidus-TcpxOut))
        return RescaledTopx

    def _Fopx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-absent melting at the given P and T. Eq(8).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-absent melting.
        """
        FcpxOut = self._FcpxOut(P, **kwargs)
        RescaledTopx = self._RescaledTopx(T, P, **kwargs)
        FopxDry = FcpxOut + (1-FcpxOut) * RescaledTopx**self.parameters['beta2']
        return FopxDry


class klb1:
    """
    Implementation of the KLB1 melting model from Matthews et al. (2021).

    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class.

    - Mcpx:   Mass fraction of cpx in the source. Controls the transition to low-productivity
      harzburgite-type melting.
    - A1:     Parameter used to define solidus.
    - A2:     Parameter used to define solidus.
    - A3:     Parameter used to define solidus.
    - A4:		Parameter used to define solidus.
    - B1:     Parameter used to define liquidus.
    - B2:     Parameter used to define liquidus.
    - B3:     Parameter used to define liquidus.
    - B4:     Parameter used to define liquidus.
    - C:     	Parameter used to define lherzolite-liquidus.
    - beta1:  Parameter used to calculate melt fraction during cpx-present melting.
    - beta2:  Parameter used to calculate melt fraction during cpx-absent melting.
    - r1:     Parameter used to define cpx reaction coefficient.
    - r2:     Parameter used to define cpx reaction coefficient.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Matthews et al. (2021)
        The model parameters described above

    """

    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'Mcpx':     0.1500,
                             'A1':    2445.7540,
                             'A2':       9.5110,
                             'A3': - 99.7820,
                             'A4': - 4378.5810,
                             'B1':     480.4030,
                             'B2':     672.3910,
                             'B3':      12.2750,
                             'B4': - 1242.5360,
                             'C':        0.6873,
                             'beta1':    1.5000,
                             'beta2':    1.5000,
                             'r1':       0.5000,
                             'r2':       0.0800
                             }
                 ):
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.DeltaS = DeltaS
        self.parameters = parameters

    def TSolidus(self, P, **kwargs):
        """
        Returns the temperature of the solidus at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperature (degC).
        """
        TSolidus = (self.parameters['A1']*_np.log(P + self.parameters['A2'])
                    + self.parameters['A3']*P + self.parameters['A4'])
        return TSolidus

    def TLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the liquidus at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperature (degC).
        """
        TLiquidus = (self.parameters['B1']*_np.log(P + self.parameters['B2'])
                     + self.parameters['B3']*P + self.parameters['B4'])
        return TLiquidus

    def F(self, P, T, **kwargs):
        """
        Wrapper for the melt fraction functions. If T and P are below the solidus,
        returns 0, if they are above the liquidus, returns 1. If below the temperature
        of cpx-exhaustion, calls the Fcpx function, otherwise calls the Fopx function.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            Melt fraction.
        """
        if T > self.TLiquidus(P, **kwargs):
            F = 1.0
        elif T < self.TSolidus(P, **kwargs):
            F = 0.0
        elif T < self._TcpxOut(P, **kwargs):
            F = self._Fcpx(T, P, **kwargs)
        else:
            F = self._Fopx(T, P, **kwargs)

        return F

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P). First calculates the melt fraction. If F is
        zero, returns _np.inf. If F is 1, returns _np.inf. Otherwise uses the
        appropriate expressions for cpx present or absent melting.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)

        Returns
        -------
        float
            dT/dF(const. P) (K).
        """
        F = self.F(P, T, **kwargs)
        if F == 0:
            # If no melt fraction the derivative is zero. Prevents division by zero.
            dTdF = _np.inf
        elif F < self._FcpxOut(P, **kwargs):
            dTdF = (((1/self.parameters['beta1']))
                    * (self._TLherzLiquidus(P, **kwargs)-self.TSolidus(P, **kwargs))
                    * (F**((1-self.parameters['beta1'])/self.parameters['beta1'])))
        elif F < 1.0:
            dTdF = (((1/self.parameters['beta2']))
                    * (self.TLiquidus(P, **kwargs)-self._TcpxOut(P, **kwargs))
                    * (F**((1-self.parameters['beta2'])/self.parameters['beta2'])))
        else:
            dTdF = _np.inf
        return dTdF

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F). First calculates F, then chooses the
        appropriate expression for cpx present or absent melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """
        F = self.F(P, T, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        dTdPLiquidus = self._dTdPLiquidus(P, **kwargs)
        dTdPLherzLiquidus = self._dTdPLherzLiquidus(P, **kwargs)
        TcpxOut = self._dTdPcpxOut(P, **kwargs)
        dTdPcpxOut = self._dTdPcpxOut(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dFdPcpxOut = self._dFdPcpxOut(P, **kwargs)

        if F == 0:
            dTdP = self.alphas/self.rhos/self.CP
        elif F < self._FcpxOut(P, **kwargs):
            dTdP = (((F**(1/self.parameters['beta1']))
                    * (dTdPLherzLiquidus-dTdPSolidus)) + dTdPSolidus)
        elif F < 1.0:
            Trel = (T - TcpxOut) / (TLiquidus - TcpxOut)
            dTdP = ((TLiquidus - TcpxOut)/(1 - FcpxOut)
                    * (1/self.parameters['beta2'])*Trel**(1-self.parameters['beta2'])
                    * dFdPcpxOut * (Trel**self.parameters['beta2']-1)
                    + dTdPcpxOut + Trel*(dTdPLiquidus - dTdPcpxOut))
        else:
            dTdP = self.alphaf/self.rhof/self.CP
        return dTdP

    def _dTdPLiquidus(self, P, **kwargs):
        """
        Returns the liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperaure gradient (degC/GPa)
        """
        dTdPLiquidus = (self.parameters['B1']/(P + self.parameters['B2'])) + self.parameters['B3']
        return dTdPLiquidus

    def _TLherzLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the lherzolite liquidus at any given pressure.
        This is the temperature at which the rock would be completely
        molten if cpx was remained present for the entirety of melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Lherzolite liquidus temperature (degC).
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        TLherzLiquidus = self.parameters['C']*TSolidus + (1 - self.parameters['C'])*TLiquidus
        return TLherzLiquidus

    def _dTdPSolidus(self, P, **kwargs):
        """
        Returns the solidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperaure gradient (degC/GPa)
        """
        dTdPSolidus = (self.parameters['A1']/(P + self.parameters['A2'])) + self.parameters['A3']
        return dTdPSolidus

    def _dTdPLherzLiquidus(self, P, **kwargs):
        """
        Returns the lherzolite liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Lherzolite liquidus temperaure gradient (degC/GPa)
        """
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        dTdPLiquidus = self._dTdPLiquidus(P, **kwargs)
        dTdPLherzoliteLiquidus = (self.parameters['C']*dTdPSolidus
                                  + (1 - self.parameters['C'])*dTdPLiquidus)
        return dTdPLherzoliteLiquidus

    def _RescaledTcpx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-present melting.

        Parameters
        ----------
        T:  float
            Temperature (degC).
        P:  float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        RescaledTemperaturecpx = ((T - TSolidus)/(TLherzLiquidus-TSolidus))
        return RescaledTemperaturecpx

    def _Fcpx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-present melting at the given P and T. Eq(2).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-present melting.
        """
        Fcpx = self._RescaledTcpx(T, P, **kwargs)**self.parameters['beta1']
        return Fcpx

    def _RxnCoef(self, P, **kwargs):
        """
        Reaction coefficient for cpx during melting at the specified pressure. Eq(7).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Reaction Coefficient.
        """
        RxnCoef = self.parameters['r1'] + self.parameters['r2']*P
        return RxnCoef

    def _FcpxOut(self, P, **kwargs):
        """
        Calculates the melt fraction required to exhaust cpx from the residue
        at the given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Melt fraction at which cpx is exhausted.
        """
        FcpxOut = self.parameters['Mcpx'] / self._RxnCoef(P, **kwargs)
        return FcpxOut

    def _dFdPcpxOut(self, P, **kwargs):
        """
        Calculates the first derivative of FcpxOut.

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            First derivative of FcpxOut.
        """
        RxnCoef = self._RxnCoef(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        return ((-1)*FcpxOut*self.parameters['r2'])/RxnCoef

    def _TcpxOut(self, P, **kwargs):
        """
        Calculates the temperature at which cpx will be exhausted during melting
        at the given pressure. Eq(9).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Temperature of cpx-exhaustion.
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        TcpxOut = (((FcpxOut**(1/self.parameters['beta1'])))
                   * (TLherzLiquidus-TSolidus) + TSolidus)
        return TcpxOut

    def _dTdPcpxOut(self, P, **kwargs):
        """
        Calculates the temperature gradient of the cpx-exhaustion surface.

        Parameters
        ----------

        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Temperature gradient of cpx-exhaustion.
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        dTdPLherzLiquidus = self._dTdPLherzLiquidus(P, **kwargs)
        dFdPcpxOut = self._dFdPcpxOut(P, **kwargs)

        A = ((dFdPcpxOut*(1/self.parameters['beta1'])
              * ((FcpxOut**((1/self.parameters['beta1'])-1))))
             * (TLherzLiquidus-TSolidus))
        B = (((FcpxOut**(1/self.parameters['beta1']))
             * (dTdPLherzLiquidus-dTdPSolidus))+dTdPSolidus)
        dTdPcpxOut = A + B
        return dTdPcpxOut

    def _RescaledTopx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-absent melting.

        Parameters
        ----------
        T:  float
            Temperature (degC).
        P:  float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TcpxOut = self._TcpxOut(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        RescaledTopx = ((T-TcpxOut)/(TLiquidus-TcpxOut))
        return RescaledTopx

    def _Fopx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-absent melting at the given P and T.

        Parameters
        ----------
        T:  float
            Temperature (degC).
        P:  float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-absent melting.
        """
        FcpxOut = self._FcpxOut(P, **kwargs)
        RescaledTopx = self._RescaledTopx(T, P, **kwargs)
        FopxDry = FcpxOut + (1-FcpxOut) * RescaledTopx**self.parameters['beta2']
        return FopxDry


class eclogite(_lithology):
    """
    Implementation of the silica-saturated pyroxenite (or eclogite) melting model from
    Matthews et al. (2021).

    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class.

    - C1:     Parameter used in solidus definition.
    - C2:     Parameter used in solidus definition.
    - C3:     Parameter used in solidus definition.
    - C4:     Parameter used in solidus definition.
    - D1:     Parameter used in liquidus definition.
    - D2:     Parameter used in liquidus definition.
    - D3:     Parameter used in liquidus definition.
    - D4:     Parameter used in liquidus definition.
    - beta:   Parameter used in melt fraction definition.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Matthews et al. (2021)
        The model parameters described above

    """

    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'C1':    533.842,
                             'C2':      4.921,
                             'C3':     20.148,
                             'C4':     80.879,
                             'D1':    994.149,
                             'D2':      8.092,
                             'D3': - 11.778,
                             'D4': - 862.641,
                             'beta':    2.134,
                             }
                 ):

        self.DeltaS = DeltaS
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.parameters = parameters

    def F(self, P, T, **kwargs):
        """
        Calculates melt fraction at a given pressure and temperature using T'**beta, where T is the
        normalised temperature: (T-Tsolidus)/(T-Tliquidus). If P and T are below the the solidus,
        0 is returned, if they are above the liquidus, 1 is returned.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)

        Returns
        -------
        float
            Melt fraction.
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        if T < Tsol:
            F = 0.0
        elif T > Tliq:
            F = 1.0
        else:
            Tr = (T-Tsol)/(Tliq-Tsol)
            F = Tr**self.parameters['beta']
        return F

    def TSolidus(self, P, **kwargs):
        """
        Calculates the solidus temperature at a given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperature (degC).
        """
        Tsol = (self.parameters['C1']
                * _np.log(P + self.parameters['C2'])
                + self.parameters['C3']*P + self.parameters['C4'])
        return Tsol

    def TLiquidus(self, P, **kwargs):
        """
        Calculates the liquidus temperature at a given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Liquidus temperature (degC).
        """
        Tliq = (self.parameters['D1'] * _np.log(P + self.parameters['D2'])
                + self.parameters['D3']*P + self.parameters['D4'])
        return Tliq

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P) at a given pressure and temperature.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dT/dF(const. P) (K)
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        F = self.F(P, T, **kwargs)
        if T < Tsol:
            dTdF = _np.inf
        elif T > Tliq:
            dTdF = _np.inf
        else:
            dTdF = ((1/self.parameters['beta']) * (Tliq-Tsol) * F**((1/self.parameters['beta'])-1))

        return dTdF

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F) at a given pressure and temperature.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dTdP(const. F) (K GPa-1)
        """
        dTdPsol = self._dTdPSolidus(P, **kwargs)
        dTdPliq = self._dTdPLiquidus(P, **kwargs)
        F = self.F(P, T, **kwargs)
        if F == 0:
            dTdP = self.alphas/self.rhos/self.CP
        elif F == 1:
            dTdP = self.alphaf/self.rhof/self.CP
        else:
            dTdP = (F**(1/self.parameters['beta'])) * (dTdPliq-dTdPsol) + dTdPsol
        return dTdP

    def _dTdPSolidus(self, P, **kwargs):
        """
        Returns the solidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperaure gradient (degC/GPa)
        """
        dTdPSolidus = ((self.parameters['C1']/(P + self.parameters['C2']))
                       + self.parameters['C3'])
        return dTdPSolidus

    def _dTdPLiquidus(self, P, **kwargs):
        """
        Returns the liquidus temperature gradient at any given pressure.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperaure gradient (degC/GPa)
        """
        dTdPLiquidus = ((self.parameters['D1']/(P + self.parameters['D2']))
                        + self.parameters['D3'])
        return dTdPLiquidus



"""
======================
Shorttle et al. (2014)
======================

Implementation of the new lithologies in Shorttle et al. (2014).

"""

from pyMelt.lithology_classes import lithology as _lithology
from pyMelt.lithology_classes import default_properties as _default_properties

import numpy as _np


class kg1(_lithology):
    """
    Implementation of the KG1 parameterisation by Shorttle et al. (2014).

    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class, with
    values:

    - A1:     Constant used in solidus expression.
    - A2:     Constant used in solidus expression.
    - A3:     Constant used in solidus expression.
    - B1:     Constant used in cpx-out expression.
    - B2:     Constant used in cpx-out expression.
    - B3:     Constant used in cpx-out expression.
    - C1:     Constant used in liquidus expression.
    - C2:     Constant used in liquidus expression.
    - C3:     Constant used in liquidus expression.
    - a:      Constant used in cpx-present melt fraction expression.
    - b:      Constant used in cpx-present melt fraction expression.
    - c:      Constant used in cpx-absent melt fraction expression.
    - d:      Constant used in cpx-absent melt fraction expression.
    - alpha:  Exponent used in the cpx-present melt fraction expression.
    - beta:   Exponent used in the cpx-absent melt fraction expression.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.lithology_class.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.lithology_class.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.lithology_class.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.lithology_class.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.lithology_class.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.lithology_class.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Shorttle et al. (2014)
        The model parameters described above
    """
    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'A1':    1095.4,
                             'A2':     124.1,
                             'A3': - 4.7,
                             'B1':    1179.6,
                             'B2':     157.2,
                             'B3': - 11.1,
                             'C1':    1780.0,
                             'C2':      45.0,
                             'C3': - 2.0,
                             'a':        0.3187864,
                             'b':        0.4154,
                             'c':        0.7341864,
                             'd':        0.2658136,
                             'alpha':    2.0,
                             'beta':     1.5
                             }
                 ):

        self.DeltaS = DeltaS
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.parameters = parameters

    def TSolidus(self, P, **kwargs):
        """
        Returns solidus temperature at a given pressure. T = A1 + A2*P + A3*P**2.

        Parameters
        ----------
        P : float, or list like
            Pressure in GPa.

        Returns
        -------
        float, or list like
            Solidus Temperature in degC.
        """
        T = self.parameters['A1'] + self.parameters['A2']*P + self.parameters['A3']*P**2
        return T

    def TLiquidus(self, P, **kwargs):
        """
        Returns liquidus temperature at a given pressure. T = C1 + C2*P + C3*P**2.

        Parameters
        ----------
        P :  float, or list like
            Pressure in GPa.

        Returns
        -------
        float, or list like
            Liquidus temperature in degC.
        """
        T = self.parameters['C1'] + self.parameters['C2']*P + self.parameters['C3']*P**2
        return T

    def dTdP(self, P, T, **kwargs):
        """
        Returns dT/dP (constant F) at a given pressure and temperature.

        Parameters
        ----------
        P : float
            Pressure in GPa.
        T : float
            Temperature in degC.

        Returns
        -------
        float
            dT/dP (constant F) in K GPa-1.
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        Tcpx = self._TCpxOut(P, **kwargs)

        if T < Tcpx:
            dTdP = (-(-(T - Tsol) / (Tcpx - Tsol) *
                    (self.parameters['B2'] + 2*self.parameters['B3']*P -
                     self.parameters['A2'] - self.parameters['A3']*2*P) -
                    self.parameters['A2'] - self.parameters['A3']*2*P))
        else:
            dTdP = (-(-(T - Tcpx) / (Tliq - Tcpx) *
                    (self.parameters['C2'] + self.parameters['C3']*2*P -
                     self.parameters['B2'] - self.parameters['B3']*2*P) -
                    self.parameters['B2'] - 2*self.parameters['B3']*P))

        return dTdP

    def dTdF(self, P, T, **kwargs):
        """
        Returns dT/dF (constant P) at a given pressure and temperature. If below
        the solidus, or above the liquidus, _np.inf is returned.

        Parameters
        ----------
        P : float
            Pressure in GPa.
        T : float
            Temperature in degC.

        Returns
        -------
        float
            dT/dF (constant P) in K.
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        Tcpx = self._TCpxOut(P, **kwargs)

        if T < Tsol:
            dTdF = _np.inf
        elif T < Tcpx:
            dTdF = ((Tcpx-Tsol) /
                    (self.parameters['b'] + self.parameters['a'] *
                     self.parameters['alpha'] *
                     ((T-Tsol)/(Tcpx-Tsol))**(self.parameters['alpha']-1)))
        elif T < Tliq:
            dTdF = ((Tliq - Tcpx)/(self.parameters['d'] * self.parameters['beta'] *
                    ((T-Tcpx)/(Tliq-Tcpx))**(self.parameters['beta']-1)))
        else:
            dTdF = _np.inf

        return dTdF

    def F(self, P, T, **kwargs):
        """
        Returns melt fraction at a given pressure and temperature. If below the
        solidus, returns 0. If above the liquidus, returns 1.

        Prior to cpx exhaustion:
        F = a*(Tr)**alpha _ b*Tr
        Tr = (T-Tsol)/(Tliq-Tsol)

        After cpx exhaustion:
        F = d*(Tr)**beta + c
        Tr = (T-Tcpx)/(Tliq-Tcpx)

        Parameters
        ----------
        P : float
            Pressure in GPa.
        T : float
            Temperature in degC.

        Returns
        -------
        float
            Melt fraction between 0 and 1.
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        Tcpx = self._TCpxOut(P, **kwargs)

        if T < Tsol:
            F = 0.0
        elif T > Tliq:
            F = 1.0
        elif T < Tcpx:
            Tf = (T-Tsol)/(Tcpx-Tsol)
            F = self.parameters['a'] * Tf**self.parameters['alpha'] + self.parameters['b']*Tf
        else:
            Tf = (T-Tcpx)/(Tliq-Tcpx)
            F = self.parameters['d'] * Tf**self.parameters['beta'] + self.parameters['c']
        return F

    def _TCpxOut(self, P, **kwargs):
        """
        Returns the temperature of cpx-exhaustion at a given pressure. T = B1 + B2*P + B3*P**2.

        Parameters
        ----------
        P :float, or list like
            Pressure in GPa.

        Returns
        -------
        float, or list like
            Cpx-exhaustion temperature in degC.
        """
        T = self.parameters['B1'] + self.parameters['B2']*P + self.parameters['B3']*P**2
        return T


class harzburgite(_lithology):
    """
    Material that does not melt, i.e. Harzburgite in Shorttle et al. (2014) and
    Matthews et al. (2016). Default thermodynamic constants are those used by
    Katz et al. (2003).

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt._default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt._default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt._default_properties['alphaf']
        Melt thermal expansivity, not used, here for consistency.
    rhos :       float, default: pyMelt._default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt._default_properties['rhof']
        Melt density, not used, here for consistency.
    DeltaS :     float, default: pyMelt._default_properties['DeltaS']
        The entropy of fusion, not used, here for consistency.
    parameters : dict, default: {}
        This model does not use any parameters, here for consistency.
    """
    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={}
                 ):
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.DeltaS = DeltaS
        self.parameters = parameters

    def F(self, P, T, **kwargs):
        """
        Melt Fraction. Returns 0.0.

        Parameters
        ----------
        P : float
            Pressure. There to maintain consistancy within lithology definitions.
        T : float
            Temperature. There to maintain consistancy within lithology definitions.

        Returns
        -------
        float
            Melt fraction will always be 0.0.
        """
        return 0.0

    def dTdF(self, P, T, **kwargs):
        """
        dTdF(constP). Returns _np.inf.

        Parameters
        ----------
        P : float
            Pressure. There to maintain consistancy within lithology definitions.
        T : float
            Temperature. There to maintain consistancy within lithology definitions.

        Returns
        -------
        numpy.inf
            The value will always be infinite.
        """
        return _np.inf

    def dTdP(self, P, T, **kwargs):
        """
        dTdP(constF). Returns 0.0.

        Parameters
        ----------
        P : float
            Pressure. There to maintain consistancy within lithology definitions.
        T : float
            Temperature. There to maintain consistancy within lithology definitions.

        Returns
        -------
        float
            The value will always be 0.0.
        """
        return 0.0

    def TSolidus(self, P, **kwargs):
        """
        Solidus temperature. Returns _np.inf.

        Parameters
        ----------
        P : float
            Pressure. There to maintain consistancy within lithology definitions.

        Returns
        -------
        numpy.inf
            The value will always be infinite.
        """
        if isinstance(P, list) or isinstance(P, _np.ndarray):
            return _np.array([_np.inf]*len(P))
        else:
            return _np.inf

    def TLiquidus(self, P, **kwargs):
        """
        Liquidus temperature. Returns _np.inf

        Parameters
        ----------
        P : float
            Pressure. There to maintain consistancy within lithology definitions.

        Returns
        -------
        numpy.inf
            The value will always be infinite.
        """
        if isinstance(P, list) or isinstance(P, _np.ndarray):
            return _np.array([_np.inf]*len(P))
        else:
            return _np.inf



"""
======
pyMelt
======

A python package for performing mantle melting calculations.

Developed and maintained by Simon Matthews (simonm@hi.is) and Kev Wong.

This module facilitates calculations for melting of heterogeneous mantle, based on empirical
parameterisations of the melting functions. The calculations assume complete thermal equilibrium
and complete chemical disequilibrium between adjacent mantle lithologies. Various tools are
provided for performing decompression melting calculations.
"""

__version__ = "0.5.1"
__author__ = "Simon Matthews and Kevin Wong"
__all__ = ['Mantle', 'hydrous_lithology']

import pyMelt.lithologies
# import pyMelt.meltingcolumn_classes
import pyMelt.mantle_class
from pyMelt.lithology_classes import hydrous_lithology


class Mantle(pyMelt.mantle_class.mantle):
    __doc__ = pyMelt.mantle_class.mantle.__doc__



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyMelt",
    version="1.931",
    author="Simon Matthews, Kevin Wong, Matthew Gleeson",
    author_email="simonwmatthews@gmail.com",
    description=("A python library for calculating the melting behaviour of Earth's mantle."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/simonwmatthews/pyMelt",
    packages=setuptools.find_packages(),
    install_requires=[
            'pandas',
            'numpy',
            'matplotlib',
            'scipy',
            ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
==================
Katz et al. (2003)
==================

Implementation of the anhydrous melting model presented by Katz et al. (2003).

"""

from pyMelt.lithology_classes import lithology as _lithology
from pyMelt.lithology_classes import default_properties as _default_properties

import numpy as _np


class lherzolite(_lithology):
    """
    Implementation of the Katz et al. (2003) anhydrous lherzolite melting model.

    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class.

    - Mcpx:   Mass fraction of cpx in the source. Controls the transition to low-productivity
      harzburgite-type melting.
    - A1:     Parameter used to define solidus.
    - A2:     Parameter used to define solidus.
    - A3:     Parameter used to define solidus.
    - B1:     Parameter used to define lherzolite-liquidus.
    - B2:     Parameter used to define lherzolite-liquidus.
    - B3:     Parameter used to define lherzolite-liquidus.
    - C1:     Parameter used to define liquidus.
    - C2:     Parameter used to define liquidus.
    - C3:     Parameter used to define liquidus.
    - beta1:  Parameter used to calculate melt fraction during cpx-present melting.
    - beta2:  Parameter used to calculate melt fraction during cpx-absent melting.
    - r1:     Parameter used to define cpx reaction coefficient.
    - r2:     Parameter used to define cpx reaction coefficient.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.lithology_class.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.lithology_class.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.lithology_class.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.lithology_class.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.lithology_class.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.lithology_class.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Katz et al. (2003)
        The model parameters described above
    """

    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'Mcpx':  0.15,
                             'A1': 1085.70,
                             'A2':  132.9,
                             'A3': - 5.1,
                             'B1': 1475.0,
                             'B2':   80.0,
                             'B3': - 3.2,
                             'C1': 1780.0,
                             'C2':   45.0,
                             'C3': - 2.0,
                             'beta1': 1.5,
                             'beta2': 1.5,
                             'r1':    0.5,
                             'r2':    0.08
                             }
                 ):
        self.DeltaS = DeltaS
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.parameters = parameters

    def TSolidus(self, P, **kwargs):
        """
        Returns the temperature of the solidus at any given pressure. Eqn(4).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Solidus temperature (degC).
        """
        TSolidus = (self.parameters['A1'] + self.parameters['A2']*P + self.parameters['A3']*(P**2))
        return TSolidus

    def TLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the liquidus at any given pressure. Eqn(10).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperature (degC).
        """
        TLiquidus = self.parameters['C1'] + self.parameters['C2']*P + self.parameters['C3']*(P**2)
        return TLiquidus

    def F(self, P, T, **kwargs):
        """
        Wrapper for the melt fraction functions. If T and P are below the solidus, returns 0, if
        they are above the liquidus, returns 1. If below the temperature of cpx-exhaustion, calls
        the Fcpx function, otherwise calls the Fopx function.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            Melt fraction.
        """
        if T > self.TLiquidus(P, **kwargs):
            F = 1.0
        elif T < self.TSolidus(P, **kwargs):
            F = 0.0
        elif T < self._TcpxOut(P, **kwargs):
            F = self._Fcpx(T, P, **kwargs)
        else:
            F = self._Fopx(T, P, **kwargs)

        return F

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P). First calculates the melt fraction. If F is zero, returns
        _np.inf. If F is 1, returns _np.inf. Otherwise uses the appropriate expressions for cpx
        present or absent melting.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)

        Returns
        -------
        float
            dT/dF(const. P) (K).
        """
        F = self.F(P, T, **kwargs)
        if F == 0:
            # If no melt fraction the derivative is zero. Prevents division by zero.
            dTdF = _np.inf
        elif F < self._FcpxOut(P, **kwargs):
            dTdF = (((1/self.parameters['beta1']))
                    * (self._TLherzLiquidus(P, **kwargs) - self.TSolidus(P, **kwargs))
                    * (F**((1-self.parameters['beta1'])/self.parameters['beta1'])))
        elif F < 1.0:
            dTdF = (((1/self.parameters['beta2']))
                    * (self.TLiquidus(P, **kwargs) - self._TcpxOut(P, **kwargs))
                    * (F**((1-self.parameters['beta2'])/self.parameters['beta2'])))
        else:
            dTdF = _np.inf
        return dTdF

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F). First calculates F, then chooses the
        appropriate expression for cpx present or absent melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dT/dP(const. F) (K GPa-1).
        """

        F = self.F(P, T, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dTdPSolidus = self._dTdPSolidus(P, **kwargs)
        TLiquidus = self.TLiquidus(P, **kwargs)
        dTdPLiquidus = self._dTdPLiquidus(P, **kwargs)
        dTdPLherzLiquidus = self._dTdPLherzLiquidus(P, **kwargs)
        TcpxOut = self._TcpxOut(P, **kwargs)
        dTdPcpxOut = self._dTdPcpxOut(P, **kwargs)
        FcpxOut = self._FcpxOut(P, **kwargs)
        dFdPcpxOut = self._dFdPcpxOut(P, **kwargs)

        if F == 0:
            dTdP = self.alphas/self.rhos/self.CP
        elif F < FcpxOut:
            dTdP = (((F**(1/self.parameters['beta1']))
                    * (dTdPLherzLiquidus-dTdPSolidus)) + dTdPSolidus)
        elif F < 1.0:
            Trel = (T - TcpxOut)/(TLiquidus - TcpxOut)
            dTdP = ((TLiquidus - TcpxOut)/(1 - FcpxOut)
                    * (1/self.parameters['beta2'])*Trel**(1-self.parameters['beta2'])
                    * dFdPcpxOut * (Trel**self.parameters['beta2'] - 1)
                    + dTdPcpxOut + Trel*(dTdPLiquidus - dTdPcpxOut))
        else:
            dTdP = self.alphaf/self.rhof/self.CP
        return dTdP

    def _TLherzLiquidus(self, P, **kwargs):
        """
        Returns the temperature of the lherzolite liquidus at any given pressure.
        Eqn(5). This is the temperature at which the rock would be completely
        molten if cpx was remained present for the entirety of melting.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        Tlzliq:   float
            Lherzolite liquidus temperature (degC).
        """
        TLherzLiquidus = (self.parameters['B1'] + self.parameters['B2']*P
                          + self.parameters['B3']*(P**2))
        return TLherzLiquidus

    def _RescaledTcpx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-present melting. (Eq 3).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TLherzLiquidus = self._TLherzLiquidus(P, **kwargs)
        RescaledTemperaturecpx = ((T - TSolidus)/(TLherzLiquidus - TSolidus))
        return RescaledTemperaturecpx

    def _Fcpx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-present melting at the given P and T. Eq(2).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-present melting.
        """
        RescaledT = self._RescaledTcpx(T, P, **kwargs)
        Fcpx = RescaledT**self.parameters['beta1']
        return Fcpx

    def _RxnCoef(self, P, **kwargs):
        """
        Reaction coefficient for cpx during melting at the specified pressure. Eq(7).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Reaction Coefficient.
        """
        RxnCoef = self.parameters['r1'] + self.parameters['r2']*P
        return RxnCoef

    def _FcpxOut(self, P, **kwargs):
        """
        Calculates the melt fraction required to exhaust cpx from the residue at the given
        pressure. Eq(6).

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Melt fraction at which cpx is exhausted.
        """
        FcpxOut = self.parameters['Mcpx'] / self._RxnCoef(P, **kwargs)
        return FcpxOut

    def _TcpxOut(self, P, **kwargs):
        """
        Calculates the temperature at which cpx will be exhausted during melting at the given
        pressure. Eq(9).

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Temperature of cpx-exhaustion.
        """
        TSolidus = self.TSolidus(P, **kwargs)
        TcpxOut = (((self._FcpxOut(P, **kwargs)**(1/self.parameters['beta1'])))
                   * (self._TLherzLiquidus(P, **kwargs) - TSolidus) + TSolidus)
        return TcpxOut

    def _RescaledTopx(self, T, P, **kwargs):
        """
        Calculates the rescaled temperature during cpx-absent melting.

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Rescaled Temperature (dimensionless).
        """
        TcpxOut = self._TcpxOut(P, **kwargs)
        RescaledTopx = ((T-TcpxOut) / (self.TLiquidus(P, **kwargs)-TcpxOut))
        return RescaledTopx

    def _Fopx(self, T, P, **kwargs):
        """
        Melt fraction during cpx-absent melting at the given P and T. Eq(8).

        Parameters
        ----------
        T : float
            Temperature (degC).
        P : float
            Pressure (degC).

        Returns
        -------
        float
            Melt fraction during cpx-absent melting.
        """
        FcpxOut = self._FcpxOut(P, **kwargs)
        FopxDry = (FcpxOut + (1-FcpxOut)
                   * self._RescaledTopx(T, P, **kwargs)**self.parameters['beta2'])
        return FopxDry

    def _dTdPSolidus(self, P, **kwargs):
        return self.parameters['A2'] + 2*self.parameters['A3']*P

    def _dTdPLiquidus(self, P, **kwargs):
        return self.parameters['C2'] + 2*self.parameters['C3']*P

    def _dTdPLherzLiquidus(self, P, **kwargs):
        return self.parameters['B2'] + 2*self.parameters['B3']*P

    def _dTdPcpxOut(self, P, **kwargs):
        term1 = ((self._TLherzLiquidus(P, **kwargs) - self.TSolidus(P, **kwargs))
                 / self.parameters['beta1']
                 * self._FcpxOut(P, **kwargs)**(1/self.parameters['beta1'] - 1))
        term2 = (self._FcpxOut(P, **kwargs)**(1/self.parameters['beta1'])
                 * (self._dTdPLherzLiquidus(P, **kwargs) - self._dTdPSolidus(P, **kwargs))
                 + self._dTdPSolidus(P, **kwargs))
        return term1 * self._dFdPcpxOut(P, **kwargs) + term2

    def _dFdPcpxOut(self, P, **kwargs):
        return - self.parameters['Mcpx'] / self._RxnCoef(P, **kwargs)**2 * self.parameters['r2']



"""
================================
Pertermann and Hirschmann (2003, JGR)
================================

The pertermann module implements the G2 model.

"""

from pyMelt.lithology_classes import lithology as _lithology
from pyMelt.lithology_classes import default_properties as _default_properties

import numpy as _np


class g2(_lithology):
    """
    Implementation of the Pertermann and Hirschmann (2003, JGR) G2 melting model.
    To use the same format of parameterisation for another lithology, the parameter values
    may be changed. They are provided as a dictionary during initialisation of the class, with
    values:

    - a:  Parameter used in calculating melt fraction.
    - b:  Parameter used in calculating melt fraction.
    - c:  Parameter used in liquidus definition.
    - d:  Parameter used in liquidus definition.
    - e:  Parameter used in solidus definition.
    - f:  Parameter used in solidus definition.

    The thermal expansivities, the heat capacity, the densities, and the entropy of fusion may
    also be changed during class initialisation.

    Parameters
    ----------
    CP :         float, default: pyMelt.lithology_class.default_properties['CP']
        The heat capacity (J K-1 kg-1)
    alphas :     float, default: pyMelt.lithology_class.default_properties['alphas']
        The thermal expansivity of the solid (1e-6 K-1)
    alphaf :     float, default: pyMelt.lithology_class.default_properties['alphaf']
        The thermal expansivity of the melt (1e-6 K-1)
    rhos :       float, default: pyMelt.lithology_class.default_properties['rhos']
        The density of the solid (kg m-3)
    rhof :       float, default: pyMelt.lithology_class.default_properties['rhof']
        The density of the melt (kg m-3)
    DeltaS :     float, default: pyMelt.lithology_class.default_properties['DeltaS']
        The entropy of fusion J K-1 kg-1
    parameters : dict, default: parameters from Pertermann and Hirschmann (2003, JGR)
        The model parameters described above
    """

    def __init__(self,
                 CP=_default_properties['CP'],
                 alphas=_default_properties['alphas'],
                 alphaf=_default_properties['alphaf'],
                 rhos=_default_properties['rhos'],
                 rhof=_default_properties['rhof'],
                 DeltaS=_default_properties['DeltaS'],
                 parameters={'a':     0.7368,
                             'b':     0.2632,
                             'c':   1175.000,
                             'd':    114.000,
                             'e':    920.000,
                             'f':    130.000
                             }
                 ):

        self.DeltaS = DeltaS
        self.CP = CP
        self.alphas = alphas
        self.alphaf = alphaf
        self.rhos = rhos
        self.rhof = rhof
        self.parameters = parameters

    def F(self, P, T, **kwargs):
        """
        Calculates melt fraction at a given pressure and temperature using Equation 1:
            a*T'**2 + b*T',
        where T is the normalised temperature (Equation 2):
            (T-Tsolidus)/(T-Tliquidus).
        If P and T are below the the solidus, 0 is returned, if they are above the liquidus, 1 is
        returned.

        Parameters
        ----------
        P : float
            Pressure (GPa)
        T : float
            Temperature (degC)

        Returns
        -------
        float
            Melt fraction.
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        if T < Tsol:
            F = 0.0
        elif T > Tliq:
            F = 1.0
        else:
            Tr = (T-Tsol)/(Tliq-Tsol)
            F = self.parameters['a']*Tr**2 + self.parameters['b']*Tr
        return F

    def TLiquidus(self, P, **kwargs):
        """
        Calculates the liquidus temperature, at a given pressure, using Equation 3:
            c + d*P.

        Parameters
        ----------
        P : float
            Pressure (GPa).

        Returns
        -------
        float
            Liquidus temperature (degC).
        """
        Tliq = self.parameters['c'] + self.parameters['d']*P
        return Tliq

    def TSolidus(self, P, **kwargs):
        """
        Calculates the solidus temperature, at a given pressure, using Equation 4:
            e + f*P.

        Parameters
        ----------
        P : float
            Pressure (GPa)

        Returns
        -------
        float
            Solidus temperature (degC).
        """
        Tsol = self.parameters['e'] + self.parameters['f']*P
        return Tsol

    def dTdF(self, P, T, **kwargs):
        """
        Calculates dT/dF(const. P) at a given pressure and temperature.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dT/dF(const. P) (K)
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        if T < Tsol:
            dTdF = _np.inf
        elif T > Tliq:
            dTdF = _np.inf
        else:
            dTdF = ((Tliq - Tsol) / (self.parameters['a'] * 2
                    * ((T - Tsol) / (Tliq - Tsol))+self.parameters['b']))

        return dTdF

    def dTdP(self, P, T, **kwargs):
        """
        Calculates dT/dP(const. F) at a given pressure and temperature.

        Parameters
        ----------
        P : float
            Pressure (GPa).
        T : float
            Temperature (degC).

        Returns
        -------
        float
            dTdP(const. F) (K GPa-1)
        """
        Tsol = self.TSolidus(P, **kwargs)
        Tliq = self.TLiquidus(P, **kwargs)
        dTdP = ((self.parameters['d'] - self.parameters['f'])
                * (T - Tsol)/(Tliq - Tsol) + self.parameters['f'])

        return dTdP
